(function() {
  var Connection, Document, MicroEvent, OpStream, connections, getConnection, open, types;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  if (typeof WEB != "undefined" && WEB !== null) {
    types || (types = exports.types);
  } else {
    OpStream = require('./opstream').OpStream;
    types = require('../types');
    MicroEvent = require('./microevent');
  }
  Document = (function() {
    function Document(stream, name, version, type, snapshot) {
      this.stream = stream;
      this.name = name;
      this.version = version;
      this.type = type;
      this.onOpReceived = __bind(this.onOpReceived, this);;
      this.tryFlushPendingOp = __bind(this.tryFlushPendingOp, this);;
      if (this.type.compose == null) {
        throw new Error('Handling types without compose() defined is not currently implemented');
      }
      this['snapshot'] = snapshot;
      this.inflightOp = null;
      this.inflightCallbacks = [];
      this.pendingOp = null;
      this.pendingCallbacks = [];
      this.serverOps = {};
      this.listeners = [];
      this['created'] = false;
      this.follow();
    }
    Document.prototype.follow = function(callback) {
      this.stream.on(this.name, 'op', this.onOpReceived);
      return this.stream.follow(this.name, this.version, __bind(function(msg) {
        if (msg['v'] !== this.version) {
          throw new Error("Expected version " + this.version + " but got " + msg['v']);
        }
        if (callback != null) {
          return callback();
        }
      }, this));
    };
    Document.prototype.unfollow = function(callback) {
      this.stream.removeListener(this.name, 'op', this.onOpReceived);
      return this.stream.unfollow(this.name, callback);
    };
    Document.prototype.tryFlushPendingOp = function() {
      if (this.inflightOp === null && this.pendingOp !== null) {
        this.inflightOp = this.pendingOp;
        this.inflightCallbacks = this.pendingCallbacks;
        this.pendingOp = null;
        this.pendingCallbacks = [];
        return this.stream.submit(this.name, this.inflightOp, this.version, __bind(function(response) {
          var callback, _i, _j, _len, _len2, _ref, _ref2;
          if (response['v'] === null) {
            _ref = this.inflightCallbacks;
            for (_i = 0, _len = _ref.length; _i < _len; _i++) {
              callback = _ref[_i];
              callback(null);
            }
            this.inflightOp = null;
            throw new Error(response['error']);
          }
          if (response['v'] !== this.version) {
            throw new Error('Invalid version from server');
          }
          this.serverOps[this.version] = this.inflightOp;
          this.version++;
          _ref2 = this.inflightCallbacks;
          for (_j = 0, _len2 = _ref2.length; _j < _len2; _j++) {
            callback = _ref2[_j];
            callback(this.inflightOp, null);
          }
          this.inflightOp = null;
          return this.tryFlushPendingOp();
        }, this));
      }
    };
    Document.prototype.onOpReceived = function(msg) {
      var docOp, op, xf, _ref, _ref2;
      if (msg['v'] < this.version) {
        return;
      }
      if (msg['doc'] !== this.name) {
        throw new Error("Expected docName " + this.name + " but got " + msg['doc']);
      }
      if (msg['v'] !== this.version) {
        throw new Error("Expected version " + this.version + " but got " + msg['v']);
      }
      op = msg['op'];
      this.serverOps[this.version] = op;
      xf = this.type.transformX || __bind(function(server, client) {
        var client_, server_;
        server_ = this.type.transform(server, client, 'server');
        client_ = this.type.transform(client, server, 'client');
        return [server_, client_];
      }, this);
      docOp = op;
      if (this.inflightOp !== null) {
        _ref = xf(docOp, this.inflightOp), docOp = _ref[0], this.inflightOp = _ref[1];
      }
      if (this.pendingOp !== null) {
        _ref2 = xf(docOp, this.pendingOp), docOp = _ref2[0], this.pendingOp = _ref2[1];
      }
      this['snapshot'] = this.type.apply(this['snapshot'], docOp);
      this.version++;
      this.emit('remoteop', docOp);
      return this.emit('change', docOp);
    };
    Document.prototype.submitOp = function(op, v, callback) {
      var realOp, _ref;
      if (v == null) {
        v = this.version;
      }
      if (typeof v === 'function') {
        callback = v;
        v = this.version;
      }
      if (((_ref = this.type) != null ? _ref.normalize : void 0) != null) {
        op = this.type.normalize(op);
      }
      while (v < this.version) {
        realOp = this.recentOps[v];
        if (!realOp) {
          throw new Error('Op version too old');
        }
        op = this.type.transform(op, realOp, 'client');
        v++;
      }
      this['snapshot'] = this.type.apply(this['snapshot'], op);
      if (this.pendingOp !== null) {
        this.pendingOp = this.type.compose(this.pendingOp, op);
      } else {
        this.pendingOp = op;
      }
      if (callback) {
        this.pendingCallbacks.push(callback);
      }
      this.emit('change', op);
      return setTimeout(this.tryFlushPendingOp, 0);
    };
    Document.prototype.close = function(callback) {
      return this.unfollow(__bind(function() {
        if (callback) {
          callback();
        }
        this.emit('closed');
      }, this));
    };
    return Document;
  })();
  MicroEvent.mixin(Document);
  Document.prototype['submitOp'] = Document.prototype.submitOp;
  Document.prototype['close'] = Document.prototype.close;
  Connection = (function() {
    function Connection(host, port, basePath) {
      this.stream = new OpStream(host, port, basePath);
      this.docs = {};
      this.numDocs = 0;
    }
    Connection.prototype.makeDoc = function(name, version, type, snapshot) {
      var doc;
      if (this.docs[name]) {
        throw new Error("Document " + name + " already followed");
      }
      doc = new Document(this.stream, name, version, type, snapshot);
      this.docs[name] = doc;
      this.numDocs++;
      doc.on('closed', __bind(function() {
        delete this.docs[name];
        return this.numDocs--;
      }, this));
      return doc;
    };
    Connection.prototype.openExisting = function(docName, callback) {
      if (this.docs[docName] != null) {
        return this.docs[docName];
      }
      return this.stream.get(docName, __bind(function(response) {
        var type;
        if (response['snapshot'] === null) {
          return callback(null);
        } else {
          type = types[response['type']];
          return callback(this.makeDoc(response['doc'], response['v'], type, response['snapshot']));
        }
      }, this));
    };
    Connection.prototype.open = function(docName, type, callback) {
      var doc;
      if (typeof type === 'function') {
        callback = type;
        type = 'text';
      }
      callback || (callback = function() {});
      if (typeof type === 'string') {
        type = types[type];
      }
      if (this.docs[docName] != null) {
        doc = this.docs[docName];
        if (doc.type === type) {
          callback(doc);
        } else {
          callback(doc, 'Document already exists with type ' + doc.type.name);
        }
        return;
      }
      return this.stream.get(docName, __bind(function(response) {
        if (response['snapshot'] === null) {
          return this.stream.submit(docName, {
            'type': type.name
          }, 0, __bind(function(response) {
            if (response['v'] != null) {
              doc = this.makeDoc(docName, 1, type, type.initialVersion());
              doc['created'] = true;
              return callback(doc);
            } else if (response['v'] === null && response['error'] === 'Type already set') {
              return this.open(docName, type, callback);
            } else {
              return callback(null, response['error']);
            }
          }, this));
        } else if (response['type'] === type.name) {
          return callback(this.makeDoc(docName, response['v'], type, response['snapshot']));
        } else {
          return callback(null, "Document already exists with type " + response['type']);
        }
      }, this));
    };
    Connection.prototype.create = function(type, prefix) {
      throw new Error('Not implemented');
    };
    Connection.prototype.disconnect = function() {
      if (this.stream != null) {
        this.emit('disconnected');
        this.stream.disconnect();
        return this.stream = null;
      }
    };
    return Connection;
  })();
  Connection.prototype['openExisting'] = Connection.prototype.openExisting;
  Connection.prototype['open'] = Connection.prototype.open;
  MicroEvent.mixin(Connection);
  connections = {};
  getConnection = function(host, port, basePath) {
    var address, c;
    if (typeof WEB != "undefined" && WEB !== null) {
      host != null ? host : host = window.location.hostname;
      port != null ? port : port = window.location.port;
    }
    address = host;
    if (port != null) {
      address += ":" + port;
    }
    if (!connections[address]) {
      c = new Connection(host, port, basePath);
      c.on('disconnected', function() {
        return delete connections[address];
      });
      connections[address] = c;
    }
    return connections[address];
  };
  open = function(docName, type, options, callback) {
    var c;
    if (typeof options === 'function') {
      callback = options;
      options = null;
    }
    options != null ? options : options = {};
    c = getConnection(options.host, options.port, options.basePath);
    return c.open(docName, type, function(doc) {
      doc.on('closed', function() {
        return setTimeout(function() {
          if (c.numDocs === 0) {
            return c.disconnect();
          }
        }, 0);
      });
      return callback(doc);
    });
  };
  if (typeof WEB != "undefined" && WEB !== null) {
    exports['Connection'] = Connection;
    exports['Document'] = Document;
    exports['open'] = open;
    window['sharejs'] = exports;
  } else {
    exports.Connection = Connection;
    exports.open = open;
  }
}).call(this);
