(function() {
  var OpStream, io;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  if (typeof WEB != "undefined" && WEB !== null) {
    if (!window['io']) {
      throw new Error('Must load socket.io before this library');
    }
    io = window['io'];
  } else {
    io = require('../../thirdparty/Socket.io-node-client').io;
  }
  OpStream = (function() {
    function OpStream(host, port, path) {
      this.onMessage = __bind(this.onMessage, this);;      var resource;
      resource = path ? path + '/socket.io' : 'socket.io';
      this.socket = new io['Socket'](host, {
        port: port,
        resource: resource
      });
      this.socket['on']('connect', this.onConnect);
      this.socket['on']('message', this.onMessage);
      this.socket['connect']();
      this.callbacks = {};
      this.lastReceivedDoc = null;
      this.lastSentDoc = null;
    }
    OpStream.prototype.onConnect = function() {};
    OpStream.prototype.on = function(docName, type, callback) {
      var _base;
      (_base = this.callbacks)[docName] || (_base[docName] = {});
      if (this.callbacks[docName][type] != null) {
        throw new Error("Callback already exists for " + docName + ", " + type);
      }
      return this.callbacks[docName][type] = callback;
    };
    OpStream.prototype.removeListener = function(docName, type, listener) {
      var _ref;
      return (_ref = this.callbacks[docName]) != null ? delete _ref[type] : void 0;
    };
    OpStream.prototype.onMessage = function(data) {
      var emit;
      if (data['doc'] != null) {
        this.lastReceivedDoc = data['doc'];
      } else {
        data['doc'] = this.lastReceivedDoc;
      }
      emit = __bind(function(type, clear) {
        var callback, _ref;
        callback = (_ref = this.callbacks[data['doc']]) != null ? _ref[type] : void 0;
        if (callback != null) {
          if (clear) {
            this.callbacks[data['doc']][type] = null;
          }
          return callback(data);
        }
      }, this);
      if (data['snapshot'] !== void 0) {
        return emit('snapshot', true);
      } else if (data['follow'] != null) {
        if (data['follow']) {
          return emit('follow', true);
        } else {
          return emit('unfollow', true);
        }
      } else if (data['v'] !== void 0) {
        if (data['op'] != null) {
          return emit('op', false);
        } else {
          return emit('localop', true);
        }
      }
    };
    OpStream.prototype.send = function(msg) {
      if (msg['doc'] === this.lastSentDoc) {
        delete msg['doc'];
      } else {
        this.lastSentDoc = msg['doc'];
      }
      return this.socket['send'](msg);
    };
    OpStream.prototype.follow = function(docName, v, callback) {
      var request;
      request = {
        'doc': docName,
        'follow': true
      };
      if (v != null) {
        request['v'] = v;
      }
      this.send(request);
      return this.on(docName, 'follow', callback);
    };
    OpStream.prototype.get = function(docName, callback) {
      this.send({
        'doc': docName,
        'snapshot': null
      });
      return this.on(docName, 'snapshot', callback);
    };
    OpStream.prototype.submit = function(docName, op, version, callback) {
      this.send({
        'doc': docName,
        'v': version,
        'op': op
      });
      return this.on(docName, 'localop', callback);
    };
    OpStream.prototype.unfollow = function(docName, callback) {
      this.send({
        'doc': docName,
        'follow': false
      });
      return this.on(docName, 'unfollow', callback);
    };
    OpStream.prototype.disconnect = function() {
      this.socket['disconnect']();
      return this.socket = null;
    };
    return OpStream;
  })();
  exports.OpStream = OpStream;
}).call(this);
