(function() {
  var defaultBasePath, expectJSONObject, getDocName, http, pump, send200, send400, send404, sendJSON, sys, url, util;
  var __hasProp = Object.prototype.hasOwnProperty;
  http = require('http');
  sys = require('sys');
  util = require('util');
  url = require('url');
  send404 = function(res, message) {
    if (message == null) {
      message = '404: Your document could not be found.\n';
    }
    res.writeHead(404, {
      'Content-Type': 'text/plain'
    });
    return res.end(message);
  };
  send400 = function(res, message) {
    res.writeHead(400, {
      'Content-Type': 'text/plain'
    });
    return res.end(message);
  };
  send200 = function(res, message) {
    if (message == null) {
      message = 'OK';
    }
    res.writeHead(200, {
      'Content-Type': 'text/plain'
    });
    return res.end(message);
  };
  getDocName = function(req) {
    var path;
    path = url.parse(req.url).pathname;
    return path.slice(1);
  };
  sendJSON = function(res, obj) {
    res.writeHead(200, {
      'Content-Type': 'application/json'
    });
    return res.end(JSON.stringify(obj) + '\n');
  };
  expectJSONObject = function(req, res, callback) {
    return pump(req, function(data) {
      var obj;
      try {
        obj = JSON.parse(data);
      } catch (error) {
        send400(res, 'Supplied JSON invalid');
        return;
      }
      return callback(obj);
    });
  };
  pump = function(req, callback) {
    var data;
    data = '';
    req.on('data', function(chunk) {
      return data += chunk;
    });
    return req.on('end', function() {
      return callback(data);
    });
  };
  defaultBasePath = '/';
  exports.attach = function(server, model, options) {
    var basePath, key, methods, router, value, _ref;
    basePath = (_ref = options != null ? options.basePath : void 0) != null ? _ref : defaultBasePath;
    for (key in defaultOptions) {
      value = defaultOptions[key];
      options[key] = value;
    }
    methods = {
      GET: function(req, res) {
        var docName;
        docName = getDocName(req);
        return model.getSnapshot(docName, function(doc) {
          if (doc.snapshot != null) {
            doc.type = doc.type.name;
            return sendJSON(res, doc);
          } else {
            return send404(res);
          }
        });
      },
      POST: function(req, res) {
        var docName, query, version;
        docName = getDocName(req);
        query = url.parse(req.url, true).query;
        version = parseInt(query != null ? query.v : void 0);
        if (version == null) {
          return send400(res, 'Version required - attach query parameter ?v=X on your URL');
        } else {
          return expectJSONObject(req, res, function(obj) {
            var opData;
            opData = {
              v: version,
              op: obj,
              meta: {
                source: req.socket.remoteAddress
              }
            };
            return model.applyOp(docName, opData, function(error, newVersion) {
              if (error != null) {
                return send400(res, error.stack);
              } else {
                return sendJSON(res, {
                  v: newVersion
                });
              }
            });
          });
        }
      },
      DELETE: function(req, res) {
        var docName;
        docName = getDocName(req);
        return model["delete"](docName, function(result) {
          if (result) {
            return send200(res);
          } else {
            return send404(res, "The document to be deleted doesn't exist");
          }
        });
      }
    };
    router = function(req, res) {
      var method;
      method = methods[req.method];
      method != null ? method : method = function(req, res) {
        var allowedMethods, k, _;
        allowedMethods = ((function() {
          var _results;
          _results = [];
          for (k in methods) {
            if (!__hasProp.call(methods, k)) continue;
            _ = methods[k];
            _results.push(k);
          }
          return _results;
        })()).join(',');
        res.writeHead(405, {
          'Content-Type': 'text/plain',
          'Allow': allowedMethods
        });
        return res.end('HTTP method not allowed');
      };
      try {
        return method(req, res);
      } catch (error) {
        res.writeHead(500, {
          'Content-Type': 'text/plain'
        });
        res.end("Internal server error:\n" + error.stack);
        return util.debug("Internal server error - " + error.stack);
      }
    };
    return server.on('request', router);
  };
}).call(this);
