(function() {
  var Events, Model, db, i, p, types;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };
  p = function() {};
  i = function() {};
  types = require('../types');
  db = require('./db');
  Events = require('./events');
  module.exports = Model = function(db, options) {
    var applyOpInternal, events, flushOps, getSnapshot, pendingOps;
    if (!(this instanceof Model)) {
      return new Model(db);
    }
    options != null ? options : options = {};
    this.getOps = db.getOps;
    this.getSnapshot = getSnapshot = function(docName, callback) {
      return db.getSnapshot(docName, function(data) {
        p("getSnapshot " + (i(data)));
        if (data.type !== null) {
          data.type = types[data.type];
        }
        return callback(data);
      });
    };
    this.getVersion = db.getVersion;
    applyOpInternal = function(docName, opData, callback) {
      p("applyOpInternal v" + opData.v + " " + (i(opData.op)) + " to " + docName + ".");
      return getSnapshot(docName, function(docData) {
        var meta, op, opVersion, snapshot, submit, type, typeName, version;
        opVersion = opData.v;
        op = opData.op;
        meta = opData.meta || {};
        meta.ts = Date.now();
        version = docData.v;
        snapshot = docData.snapshot;
        type = docData.type;
        p("applyOp hasdata v" + opVersion + " " + (i(op)) + " to " + docName + ".");
        submit = function() {
          var newDocData, newOpData, _ref;
          newOpData = {
            op: op,
            v: opVersion,
            meta: meta
          };
          newDocData = {
            snapshot: snapshot,
            type: (_ref = type != null ? type.name : void 0) != null ? _ref : null
          };
          p("submit " + (i(newOpData)));
          return db.append(docName, newOpData, newDocData, function() {
            p("appended v" + opVersion + " to " + docName + ". Calling callback...");
            events.onApplyOp(docName, newOpData);
            return callback(null, opVersion);
          });
        };
        if (opVersion > version) {
          callback(new Error('Op at future version'), null);
          return;
        }
        if (opVersion === 0) {
          if (version === 0) {
            typeName = op.type;
            if (typeName == null) {
              callback(new Error('Invalid op: type required'), null);
              return;
            }
            type = types[typeName];
            if (type == null) {
              callback(new Error("Invalid op: type '" + typeName + "' missing"), null);
              return;
            }
            snapshot = type.initialVersion();
            return submit();
          } else {
            callback(new Error('Type already set'), null);
          }
        } else {
          if (opVersion < version) {
            return db.getOps(docName, opVersion, version, function(ops) {
              var realOp, _i, _len;
              try {
                for (_i = 0, _len = ops.length; _i < _len; _i++) {
                  realOp = ops[_i];
                  p("XFORM Doc " + docName + " op " + (i(op)) + " by " + (i(realOp.op)));
                  op = docData.type.transform(op, realOp.op, 'client');
                  opVersion++;
                  p("-> " + (i(op)));
                }
                snapshot = docData.type.apply(docData.snapshot, op);
              } catch (error) {
                callback(error, null);
                return;
              }
              return submit();
            });
          } else {
            try {
              snapshot = docData.type.apply(docData.snapshot, op);
            } catch (error) {
              callback(error, null);
              return;
            }
            return submit();
          }
        }
      });
    };
    pendingOps = {};
    flushOps = function(docName) {
      var callback, opData, state, _ref;
      state = pendingOps[docName];
      p("flushOps " + docName + " state " + (i(state)));
      if (state.busy || state.queue.length === 0) {
        return;
      }
      p("continuing...");
      state.busy = true;
      _ref = state.queue.shift(), opData = _ref[0], callback = _ref[1];
      return applyOpInternal(docName, opData, function(error, version) {
        if (callback != null) {
          callback(error, version);
        }
        state.busy = false;
        return flushOps(docName);
      });
    };
    this.applyOp = function(docName, opData, callback) {
      p("applyOp " + docName + " op " + (i(opData)));
      pendingOps[docName] || (pendingOps[docName] = {
        busy: false,
        queue: []
      });
      pendingOps[docName].queue.push([opData, callback]);
      return flushOps(docName);
    };
    this["delete"] = function(docName, callback) {
      events.removeAllListeners(docName);
      return db["delete"](docName, callback);
    };
    events = new Events(this);
    this.listen = events.listen;
    this.removeListener = events.removeListener;
    this.listenFromVersion = events.listenFromVersion;
    this.randomDocName = function(length) {
      var chars, x;
      if (length == null) {
        length = 10;
      }
      chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-=";
      return ((function() {
        var _results;
        _results = [];
        for (x = 0; (0 <= length ? x < length : x > length); (0 <= length ? x += 1 : x -= 1)) {
          _results.push(chars[Math.floor(Math.random() * chars.length)]);
        }
        return _results;
      })()).join('');
    };
    options.auth || (options.auth = function() {});
    options.canRead || (options.canRead = function(client, docName, result) {
      return result.accept();
    });
    options.canSubmitOp || (options.canSubmitOp = function(client, docName, opData, result) {
      return result.accept();
    });
    options.canDelete || (options.canDelete = function(client, docName, result) {
      return result.reject();
    });
    this.auth = options.auth;
    this.clientGetOps = function(client, docName, start, end, callback) {
      return options.canRead(client, docName, {
        accept: __bind(function() {
          return this.getOps(docName, start, end, function(data) {
            return callback(null, data);
          });
        }, this),
        reject: function() {
          return callback(new Error('Forbidden'));
        }
      });
    };
    this.clientGetSnapshot = function(client, docName, callback) {
      return options.canRead(client, docName, {
        accept: __bind(function() {
          return this.getSnapshot(docName, function(data) {
            return callback(null, data);
          });
        }, this),
        reject: function() {
          return callback(new Error("Forbidden"));
        }
      });
    };
    this.clientSubmitOp = __bind(function(client, docName, opData, callback) {
      return options.canSubmitOp(client, docName, opData, {
        accept: function() {
          return this.applyOp(docName, opData, callback);
        },
        reject: function() {
          return callback(new Error('Forbidden'));
        }
      });
    }, this);
    this.clientDelete = function(client, docName, callback) {
      return options.canDelete(client, docName, {
        accept: function() {
          return this["delete"](docName, function(result) {
            return callback(null, result);
          });
        },
        reject: function() {
          return callback(new Error('Forbidden'));
        }
      });
    };
    return this;
  };
}).call(this);
