(function() {
  var i, io, p, util;
  io = require('socket.io');
  util = require('util');
  p = function() {};
  i = function() {};
  exports.attach = function(server, model, options) {
    var socket;
    socket = io.listen(server, {
      log: null
    });
    socket.on('connection', function(client) {
      var docState, flush, follow, lastReceivedDoc, lastSentDoc, messageListener, opReceived, send, snapshotRequest, unfollow;
      if (client.request != null) {
        p("New client connected from " + client.request.socket.remoteAddress + " with sessionId " + client.sessionId);
      }
      lastSentDoc = null;
      lastReceivedDoc = null;
      docState = {};
      send = function(msg) {
        p("Sending " + (i(msg)));
        if (msg.doc === lastSentDoc) {
          delete msg.doc;
        } else {
          lastSentDoc = msg.doc;
        }
        return client.send(msg);
      };
      follow = function(data, callback) {
        var docName, listener, sendOpenConfirmation, version;
        docName = data.doc;
        version = data.v;
        if (docState[docName].listener != null) {
          throw new Error('Doc already followed');
        }
        p("Registering follower on " + docName + " by " + client.sessionId + " at " + version);
        sendOpenConfirmation = function(v) {
          p("Following " + docName + " at " + v + " by " + client.sessionId);
          send({
            doc: docName,
            follow: true,
            v: v
          });
          return callback();
        };
        docState[docName].listener = listener = function(opData) {
          var opMsg, _ref, _ref2;
          if (docState[docName].listener !== listener) {
            throw new Error('Consistency violation - doc listener invalid');
          }
          p("follow listener doc:" + docName + " opdata:" + (i(opData)) + " v:" + version);
          if ((((_ref2 = opData.meta) != null ? _ref2.source : void 0) === (_ref = client.sessionId) && _ref !== void 0)) {
            return;
          }
          opMsg = {
            doc: docName,
            op: opData.op,
            v: opData.v,
            meta: opData.meta
          };
          return send(opMsg);
        };
        if (version != null) {
          return model.listenFromVersion(docName, version, listener, function() {
            return sendOpenConfirmation(version);
          });
        } else {
          return model.listen(docName, sendOpenConfirmation, listener);
        }
      };
      unfollow = function(data, callback) {
        var listener;
        p("Closing " + data.doc);
        listener = docState[data.doc].listener;
        if (listener == null) {
          throw new Error('Doc already closed');
        }
        model.removeListener(data.doc, listener);
        docState[data.doc].listener = null;
        send({
          doc: data.doc,
          follow: false
        });
        return callback();
      };
      opReceived = function(data, callback) {
        var op_data;
        if (data.doc == null) {
          throw new Error('No docName specified');
        }
        if (data.v == null) {
          throw new Error('No version specified');
        }
        op_data = {
          v: data.v,
          op: data.op
        };
        op_data.meta = data.meta || {};
        op_data.meta.source = client.sessionId;
        return model.applyOp(data.doc, op_data, function(error, appliedVersion) {
          var msg;
          msg = error != null ? (p("Sending error to client: " + error.message + ", " + error.stack), {
            doc: data.doc,
            v: null,
            error: error.message
          }) : {
            doc: data.doc,
            v: appliedVersion
          };
          send(msg);
          return callback();
        });
      };
      snapshotRequest = function(data, callback) {
        if (data.doc == null) {
          throw new Error('No docName specified');
        }
        return model.getSnapshot(data.doc, function(doc) {
          var msg, _ref;
          msg = {
            doc: data.doc,
            v: doc.v,
            type: ((_ref = doc.type) != null ? _ref.name : void 0) || null,
            snapshot: doc.snapshot
          };
          send(msg);
          return callback();
        });
      };
      flush = function(state) {
        var callback, data;
        p("flush state " + (i(state)));
        if (state.busy || state.queue.length === 0) {
          return;
        }
        state.busy = true;
        data = state.queue.shift();
        callback = function() {
          p('flush complete...');
          state.busy = false;
          return flush(state);
        };
        p("processing data " + (i(data)));
        try {
          if (data.follow != null) {
            if (data.follow) {
              return follow(data, callback);
            } else {
              return unfollow(data, callback);
            }
          } else if (data.op != null) {
            return opReceived(data, callback);
          } else if (data.snapshot !== void 0) {
            return snapshotRequest(data, callback);
          } else {
            return util.debug("Unknown message received: " + (util.inspect(data)));
          }
        } catch (error) {
          util.debug(error.stack);
          return callback();
        }
      };
      messageListener = function(data) {
        var _name;
        if (docState === null) {
          console.log("WARNING: received data from client after the client disconnected.");
          console.log(client);
          return;
        }
        p('on MESSAGE ' + i(data));
        try {
          if (typeof data === 'string') {
            data = JSON.parse(data);
          }
          if (data.doc === null) {
            lastReceivedDoc = null;
            data.doc = model.randomDocName();
          } else if (data.doc != null) {
            lastReceivedDoc = data.doc;
          } else {
            if (!lastReceivedDoc) {
              throw new Error('msg.doc missing');
            }
            data.doc = lastReceivedDoc;
          }
        } catch (error) {
          util.debug(error.stack);
          return;
        }
        docState[_name = data.doc] || (docState[_name] = {
          listener: null,
          queue: [],
          busy: false
        });
        docState[data.doc].queue.push(data);
        return flush(docState[data.doc]);
      };
      client.on('message', messageListener);
      return client.on('disconnect', function() {
        var docName, state;
        p("client " + client.sessionId + " disconnected");
        for (docName in docState) {
          state = docState[docName];
          state.busy = true;
          state.queue = [];
          if (state.listener != null) {
            model.removeListener(docName, state.listener);
          }
        }
        client.removeListener('message', messageListener);
        return docState = null;
      });
    });
    return server;
  };
}).call(this);
