(function() {
  var append, checkValidComponent, checkValidOp, compose, compress, invertComponent, strInject, text, transformComponent, transformComponentX, transformPosition, transformX;
  text = {};
  text.name = 'text';
  text.initialVersion = function() {
    return '';
  };
  strInject = function(s1, pos, s2) {
    return s1.slice(0, pos) + s2 + s1.slice(pos);
  };
  checkValidComponent = function(c) {
    var d_type, i_type;
    if (typeof c['p'] !== 'number') {
      throw new Error('component missing position field');
    }
    i_type = typeof c['i'];
    d_type = typeof c['d'];
    if (!((i_type === 'string') ^ (d_type === 'string'))) {
      throw new Error('component needs an i or d field');
    }
    if (!(c['p'] >= 0)) {
      throw new Error('position cannot be negative');
    }
  };
  checkValidOp = function(op) {
    var c, _i, _len;
    for (_i = 0, _len = op.length; _i < _len; _i++) {
      c = op[_i];
      checkValidComponent(c);
    }
    return true;
  };
  text.apply = function(snapshot, op) {
    var component, deleted, _i, _len;
    checkValidOp(op);
    for (_i = 0, _len = op.length; _i < _len; _i++) {
      component = op[_i];
      if (component['i'] != null) {
        snapshot = strInject(snapshot, component['p'], component['i']);
      } else {
        deleted = snapshot.slice(component['p'], component['p'] + component['d'].length);
        if (component['d'] !== deleted) {
          throw new Error("Delete component '" + component['d'] + "' does not match deleted text '" + deleted + "'");
        }
        snapshot = snapshot.slice(0, component['p']) + snapshot.slice(component['p'] + component['d'].length);
      }
    }
    return snapshot;
  };
  text._append = append = function(newOp, c) {
    var last, _ref, _ref2;
    if (c['i'] === '' || c['d'] === '') {
      return;
    }
    if (newOp.length === 0) {
      return newOp.push(c);
    } else {
      last = newOp[newOp.length - 1];
      if ((last['i'] != null) && (c['i'] != null) && (last['p'] <= (_ref = c['p']) && _ref <= (last['p'] + last['i'].length))) {
        return newOp[newOp.length - 1] = {
          'i': strInject(last['i'], c['p'] - last['p'], c['i']),
          'p': last['p']
        };
      } else if ((last['d'] != null) && (c['d'] != null) && (c['p'] <= (_ref2 = last['p']) && _ref2 <= (c['p'] + c['d'].length))) {
        return newOp[newOp.length - 1] = {
          'd': strInject(c['d'], last['p'] - c['p'], last['d']),
          'p': c['p']
        };
      } else {
        return newOp.push(c);
      }
    }
  };
  text.compose = compose = function(op1, op2) {
    var c, newOp, _i, _len;
    checkValidOp(op1);
    checkValidOp(op2);
    newOp = op1.slice();
    for (_i = 0, _len = op2.length; _i < _len; _i++) {
      c = op2[_i];
      append(newOp, c);
    }
    checkValidOp(newOp);
    return newOp;
  };
  text.compress = compress = function(op) {
    return compose([], op);
  };
  text.normalize = function(op) {
    var c, newOp, _i, _len, _ref;
    newOp = [];
    if ((op['i'] != null) || (op['p'] != null)) {
      op = [op];
    }
    for (_i = 0, _len = op.length; _i < _len; _i++) {
      c = op[_i];
      (_ref = c['p']) != null ? _ref : c['p'] = 0;
      append(newOp, c);
    }
    return newOp;
  };
  transformPosition = function(pos, c, insertAfter) {
    if (c['i'] != null) {
      if (c['p'] < pos || (c['p'] === pos && insertAfter)) {
        return pos + c['i'].length;
      } else {
        return pos;
      }
    } else {
      if (pos <= c['p']) {
        return pos;
      } else if (pos <= c['p'] + c['d'].length) {
        return c['p'];
      } else {
        return pos - c['d'].length;
      }
    }
  };
  text.transformCursor = function(position, op, insertAfter) {
    var c, _i, _len;
    for (_i = 0, _len = op.length; _i < _len; _i++) {
      c = op[_i];
      position = transformPosition(position, c, insertAfter);
    }
    return position;
  };
  transformComponent = function(dest, c, otherC, type) {
    var cIntersect, intersectEnd, intersectStart, newC, otherIntersect, s;
    checkValidOp([c]);
    checkValidOp([otherC]);
    if (c['i'] != null) {
      return append(dest, {
        'i': c['i'],
        'p': transformPosition(c['p'], otherC, type === 'server')
      });
    } else {
      if (otherC['i'] != null) {
        s = c['d'];
        if (c['p'] < otherC['p']) {
          append(dest, {
            'd': s.slice(0, otherC['p'] - c['p']),
            'p': c['p']
          });
          s = s.slice(otherC['p'] - c['p']);
        }
        if (s !== '') {
          return append(dest, {
            'd': s,
            'p': c['p'] + otherC['i'].length
          });
        }
      } else {
        if (c['p'] >= otherC['p'] + otherC['d'].length) {
          return append(dest, {
            'd': c['d'],
            'p': c['p'] - otherC['d'].length
          });
        } else if (c['p'] + c['d'].length <= otherC['p']) {
          return append(dest, c);
        } else {
          newC = {
            'd': '',
            'p': c['p']
          };
          if (c['p'] < otherC['p']) {
            newC['d'] = c['d'].slice(0, otherC['p'] - c['p']);
          }
          if (c['p'] + c['d'].length > otherC['p'] + otherC['d'].length) {
            newC['d'] += c['d'].slice(otherC['p'] + otherC['d'].length - c['p']);
          }
          intersectStart = Math.max(c['p'], otherC['p']);
          intersectEnd = Math.min(c['p'] + c['d'].length, otherC['p'] + otherC['d'].length);
          cIntersect = c['d'].slice(intersectStart - c['p'], intersectEnd - c['p']);
          otherIntersect = otherC['d'].slice(intersectStart - otherC['p'], intersectEnd - otherC['p']);
          if (cIntersect !== otherIntersect) {
            throw new Error('Delete ops delete different text in the same region of the document');
          }
          if (newC['d'] !== '') {
            newC['p'] = transformPosition(newC['p'], otherC);
            return append(dest, newC);
          }
        }
      }
    }
  };
  transformComponentX = function(server, client, destServer, destClient) {
    transformComponent(destServer, server, client, 'server');
    return transformComponent(destClient, client, server, 'client');
  };
  text.transformX = transformX = function(serverOp, clientOp) {
    var c, c_, clientComponent, k, newClientOp, newServerOp, nextC, s, s_, _i, _j, _k, _l, _len, _len2, _len3, _len4, _ref, _ref2;
    checkValidOp(serverOp);
    checkValidOp(clientOp);
    newClientOp = [];
    for (_i = 0, _len = clientOp.length; _i < _len; _i++) {
      clientComponent = clientOp[_i];
      newServerOp = [];
      k = 0;
      while (k < serverOp.length) {
        nextC = [];
        transformComponentX(serverOp[k], clientComponent, newServerOp, nextC);
        k++;
        if (nextC.length === 1) {
          clientComponent = nextC[0];
        } else if (nextC.length === 0) {
          _ref = serverOp.slice(k);
          for (_j = 0, _len2 = _ref.length; _j < _len2; _j++) {
            s = _ref[_j];
            append(newServerOp, s);
          }
          clientComponent = null;
          break;
        } else {
          _ref2 = transformX(serverOp.slice(k), nextC), s_ = _ref2[0], c_ = _ref2[1];
          for (_k = 0, _len3 = s_.length; _k < _len3; _k++) {
            s = s_[_k];
            append(newServerOp, s);
          }
          for (_l = 0, _len4 = c_.length; _l < _len4; _l++) {
            c = c_[_l];
            append(newClientOp, c);
          }
          clientComponent = null;
          break;
        }
      }
      if (clientComponent != null) {
        append(newClientOp, clientComponent);
      }
      serverOp = newServerOp;
    }
    return [serverOp, newClientOp];
  };
  text.transform = function(op, otherOp, type) {
    var client, server, _, _ref, _ref2;
    if (!(type === 'server' || type === 'client')) {
      throw new Error("type must be 'server' or 'client'");
    }
    if (type === 'server') {
      _ref = transformX(op, otherOp), server = _ref[0], _ = _ref[1];
      return server;
    } else {
      _ref2 = transformX(otherOp, op), _ = _ref2[0], client = _ref2[1];
      return client;
    }
  };
  invertComponent = function(c) {
    if (c['i'] != null) {
      return {
        'd': c['i'],
        'p': c['p']
      };
    } else {
      return {
        'i': c['d'],
        'p': c['p']
      };
    }
  };
  text.invert = function(op) {
    var c, _i, _len, _ref, _results;
    _ref = op.slice().reverse();
    _results = [];
    for (_i = 0, _len = _ref.length; _i < _len; _i++) {
      c = _ref[_i];
      _results.push(invertComponent(c));
    }
    return _results;
  };
  if (typeof WEB != "undefined" && WEB !== null) {
    exports.types || (exports.types = {});
    exports.types['text'] = text;
  } else {
    module.exports = text;
  }
}).call(this);
