(function() {
  var fs, i, p, randomInt, randomWord, text, util;
  fs = require('fs');
  util = require('util');
  randomWord = require('./randomWord');
  text = require('../../src/types/text-composable');
  p = util.debug;
  i = util.inspect;
  exports.testTransforms = function(test) {
    var expected, op, otherOp, result, testData, type;
    testData = fs.readFileSync(__dirname + '/text-transform-tests.json').toString().split('\n');
    while (testData.length >= 4) {
      op = JSON.parse(testData.shift());
      otherOp = JSON.parse(testData.shift());
      type = testData.shift();
      expected = JSON.parse(testData.shift());
      result = text.transform(op, otherOp, type);
      test.deepEqual(result, expected);
    }
    return test.done();
  };
  exports.testCompose = function(test) {
    var op1, op2, result, testData;
    testData = fs.readFileSync(__dirname + '/text-transform-tests.json').toString().split('\n');
    while (testData.length >= 4) {
      testData.shift();
      op1 = JSON.parse(testData.shift());
      testData.shift();
      op2 = JSON.parse(testData.shift());
      result = text.compose(op1, op2);
    }
    return test.done();
  };
  exports.testNormalize = function(test) {
    test.deepEqual([], text.normalize([0]));
    test.deepEqual([], text.normalize([
      {
        i: ''
      }
    ]));
    test.deepEqual([], text.normalize([
      {
        d: ''
      }
    ]));
    test.deepEqual([2], text.normalize([1, 1]));
    test.deepEqual([2], text.normalize([2, 0]));
    test.deepEqual([
      {
        i: 'a'
      }
    ], text.normalize([
      {
        i: 'a'
      }, 0
    ]));
    test.deepEqual([
      {
        i: 'ab'
      }
    ], text.normalize([
      {
        i: 'a'
      }, {
        i: 'b'
      }
    ]));
    test.deepEqual([
      {
        i: 'ab'
      }
    ], text.normalize([
      {
        i: 'ab'
      }, {
        i: ''
      }
    ]));
    test.deepEqual([
      {
        i: 'ab'
      }
    ], text.normalize([
      0, {
        i: 'a'
      }, 0, {
        i: 'b'
      }, 0
    ]));
    test.deepEqual([
      {
        i: 'a'
      }, 1, {
        i: 'b'
      }
    ], text.normalize([
      {
        i: 'a'
      }, 1, {
        i: 'b'
      }
    ]));
    return test.done();
  };
  randomInt = function(n) {
    return Math.floor(Math.random() * n);
  };
  text.generateRandomOp = function(docStr) {
    var addDelete, addInsert, addSkip, append, chance, expectedDoc, initial, op;
    initial = docStr;
    op = [];
    expectedDoc = '';
    append = text._makeAppend(op);
    addSkip = function() {
      var length;
      length = randomInt(Math.min(docStr.length, 3)) + 1;
      append(length);
      expectedDoc += docStr.slice(0, length);
      return docStr = docStr.slice(length);
    };
    addInsert = function() {
      var word;
      word = randomWord() + ' ';
      append({
        i: word
      });
      return expectedDoc += word;
    };
    addDelete = function() {
      var deletedStr, length;
      length = randomInt(Math.min(docStr.length, 7)) + 1;
      deletedStr = docStr.slice(0, length);
      append({
        d: deletedStr
      });
      return docStr = docStr.slice(length);
    };
    while (docStr.length > 0) {
      chance = initial.length > 100 ? 4 : 3;
      switch (randomInt(chance)) {
        case 0:
          addSkip();
          break;
        case 1:
          addInsert();
          break;
        case 2:
        case 3:
          addDelete();
      }
    }
    if (randomInt(3) === 0) {
      addInsert();
    }
    return [op, expectedDoc];
  };
  text.generateRandomDoc = randomWord;
  exports.randomizer = function(test) {
    require('../helpers').randomizerTest(text);
    return test.done();
  };
}).call(this);
